# The icMRCCSD(T) potential energy surface for the FHCl system
# See calc_PES.py for an example on the usage
#
import math

# The number of linear coefficients
def get_n_coef( m):
    n = 0
    for i in range(0, m+1):
        for j in range(0, m+1):
            for k in range(0, m+1):
                if (i + j + k <= m):
                    if (i*j >0 or j*k >0 or k*i >0):
                        n += 1
    return n

# Polynomial orders
M = 10
L = 0
n_coef = get_n_coef( M)
M_TS = 7
L_TS = 2
n_coef_TS = get_n_coef( M_TS)

# Equilibrium distances, icMRCCSD(T)/aVTZ
R_FH_0  = 1.741284
R_HCl_0 = 2.414313
R_ClF_0 = 3.104319
R_FH_0_TS  = 1.741284
R_HCl_0_TS = 2.414313
R_ClF_0_TS = 3.104319

# TS geometry, in perimetric coordinates
TS_rF  = 2.529145
TS_rH  = 0.493715
TS_rCl = 1.985675

# The three-body scaling factor
alpha = 1.02403


# =====================
# Two body term

# Coef[0] = beta_1
# Coef[1] = beta_2
# Coef[2] = C_0
# Coef[3] = C_1
# ...
# Coef[i] = C_{i-2} 
# ...
# Coef[N+2] = C_N

# --------
def get_coef_2b_from_file( file_name):
    f = open( file_name, 'r')
    coef = []
    for l in f:
        line_spl = l.split()
        coef.append( float( line_spl[0]))
    f.close()
    return coef

# --------
def V2( R, Coef):
    rho = R*math.exp( -Coef[1]*R)
    rho_i = rho
    S = Coef[2]*math.exp( -Coef[0]*R)/R
    for c_i in Coef[3:]:
        S += c_i*rho_i
        rho_i *= rho
    return S



# =====================
# Three body term

# Coef[0] = beta_FH
# Coef[1] = beta_ClF
# Coef[2] = beta_HCl
# Coef[3] = beta_TS_FH
# Coef[4] = beta_TS_ClF
# Coef[5] = beta_TS_HCl

# Coef[5+ 1] =        C_1      (D_0110)
# ...
# Coef[5+ n_coef] =   C_N
# Coef[5+ n_coef+1] = C_1      (D_0111)
# ...
# Coef[5+ 2*n_coef] = C_N
# ...
# ...
# Coef[5+ L*n_coef+1] = C_1      (D_011L)
# ...
# Coef[5+ (L+1)*n_coef] = C_N

# N = 5+ (L+1)*n_coef

# Coef[N+ 1] =        C_1      (D_0110)
# ...
# Coef[N+ n_coef_TS] =   C_N
# Coef[N+ n_coef_TS+1] = C_1      (D_0111)
# ...
# Coef[N+ 2*n_coef_TS] = C_N
# ...
# ...
# Coef[N+ L_TS*n_coef_TS+1] = C_1      (D_011L)
# ...
# Coef[N+ (L_TS+1)*n_coef_TS] = C_N

# betas = [beta_FH, beta_ClF, beta_HCl, beta_FH_TS, beta_ClF_TS, beta_HCl_TS]

# --------
def get_coef_3b_from_file( file_name):
    f = open( file_name, 'r')
    coef = []
    for l in f:
        line_spl = l.split()
        coef.append( float( line_spl[0]))
    f.close()
    return coef

# --------
def get_rho_contrib( R_FH, R_HCl, R_ClF, betas):
    rF  = (R_FH - R_HCl + R_ClF)/2
    rH  = (R_FH + R_HCl - R_ClF)/2
    rCl =(-R_FH + R_HCl + R_ClF)/2

    sum_cos  = (R_FH **2)*rCl
    sum_cos += (R_HCl**2)*rF
    sum_cos += (R_ClF**2)*rH
    sum_cos = 1.0 + sum_cos/(R_FH*R_HCl*R_ClF)
    a_pt = []

    # x is the contribution of TS
    dist = math.sqrt((rF-TS_rF)**2 + (rH-TS_rH)**2 + (rCl-TS_rCl)**2)
    x = math.exp(-dist**2)

    if (x < 1.0): # long range (LR)
        rho_FH  = math.exp( -betas[0]*(R_FH  - R_FH_0))
        rho_HCl = math.exp( -betas[2]*(R_HCl - R_HCl_0))
        rho_ClF = math.exp( -betas[1]*(R_ClF - R_ClF_0))
        rho_FH  = math.exp( -R_FH_0 /R_FH )*rho_FH 
        rho_HCl = math.exp( -R_HCl_0/R_HCl)*rho_HCl
        rho_ClF = math.exp( -R_ClF_0/R_ClF)*rho_ClF
        for l in range(0, L+1):
            sum_cos_l = sum_cos**l
            i = j = k = 0
            rho_FH_i = rho_ClF_j = rho_HCl_k = 1.0
            for i_m in range(0, n_coef):
                k += 1
                rho_HCl_k = rho_HCl_k*rho_HCl
                if (i + j + k > M):
                    k = 0
                    rho_HCl_k = 1.0
                    j += 1
                    rho_ClF_j = rho_ClF_j*rho_ClF
                if ((j+k == 0) or (i+k == 0)):
                    k = 1
                    rho_HCl_k = rho_HCl
                if (i + j + k > M):
                    j = 0
                    rho_ClF_j = 1.0
                    i += 1
                    rho_FH_i = rho_FH_i*rho_FH
                if ((j+k == 0) or (i+k == 0)):
                    k = 1
                    rho_HCl_k = rho_HCl
                if ((j+i == 0)):
                    j = 1
                    rho_ClF_j = rho_ClF
                a_pt.append( rho_FH_i*rho_ClF_j*rho_HCl_k*sum_cos_l*(1.0-x))
    else:
        a_pt.extend([0.0]*((L+1)*n_coef))

    if (x > 0.0): # TS region
        rho_FH  = math.exp( -betas[3]*(R_FH  - R_FH_0_TS))
        rho_HCl = math.exp( -betas[5]*(R_HCl - R_HCl_0_TS))
        rho_ClF = math.exp( -betas[4]*(R_ClF - R_ClF_0_TS))
        rho_FH  = math.exp( -R_FH_0_TS /R_FH )*rho_FH 
        rho_HCl = math.exp( -R_HCl_0_TS/R_HCl)*rho_HCl
        rho_ClF = math.exp( -R_ClF_0_TS/R_ClF)*rho_ClF
        for l in range(0, L_TS+1):
            sum_cos_l = sum_cos**l
            i = j = k = 0
            rho_FH_i = rho_ClF_j = rho_HCl_k = 1.0
            for i_m in range(0, n_coef_TS):
                k += 1
                rho_HCl_k = rho_HCl_k*rho_HCl
                if (i + j + k > M_TS):
                    k = 0
                    rho_HCl_k = 1.0
                    j += 1
                    rho_ClF_j = rho_ClF_j*rho_ClF
                if ((j+k == 0) or (i+k == 0)):
                    k = 1
                    rho_HCl_k = rho_HCl
                if (i + j + k > M_TS):
                    j = 0
                    rho_ClF_j = 1.0
                    i += 1
                    rho_FH_i = rho_FH_i*rho_FH
                if ((j+k == 0) or (i+k == 0)):
                    k = 1
                    rho_HCl_k = rho_HCl
                if ((j+i == 0)):
                    j = 1
                    rho_ClF_j = rho_ClF
                a_pt.append( rho_FH_i*rho_ClF_j*rho_HCl_k*sum_cos_l*x)
    else:
        a_pt.extend([0.0]*((L_TS+1)*n_coef_TS))
    
    return a_pt

# --------
def V3( R_FH, R_HCl, R_ClF, Coef):
    S = 0.0
    rhos = get_rho_contrib( R_FH, R_HCl, R_ClF, Coef[0:6])
    for i in range(0, (L+1)*n_coef + (L_TS+1)*n_coef_TS):
        S += Coef[6 + i]*rhos[i]
    return S



# =========
# Calculate the PES on a point and return it in hartree
# =========
def V( R_FH, R_HCl, R_ClF, coef_FH, coef_HCl, coef_ClF, coef_three):
    if (R_FH < 1.0 or R_HCl < 1.5 or R_ClF < 2.25):
        return 100.0
    V_FH  = V2( R_FH,  coef_FH)
    V_HCl = V2( R_HCl, coef_HCl)
    V_ClF = V2( R_ClF, coef_ClF)
    V_3 = V3( R_FH, R_HCl, R_ClF, coef_three)
    V_3 *= alpha
    return V_FH + V_HCl + V_ClF + V_3
